/*
 * 5 kV digitally controlled power supply with current and resistance measurements (isolation test) 
 * Firmware version 0.1.0
 * 
 * Using ATmega328p, programming through ISP. Fuses: E:FD H:DE L:FF.
 * Arduino Nano with Mega328 and upload over ISP can be used inside Arduino IDE
 *
 * License of this file (check used libraries for their respective licenses):
 *
 * Copyright 2020 F. Stefanec
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Controls: UP, DOWN: Obvious
 * SEL: Switch between set voltage/actual voltage and current/resistance displays in normal mode. Hold at start to enter calibration mode, confirm in calibration mode.
 */

#include <LiquidCrystal.h>
#include <EEPROM.h>

#define FW_VER_STRING "FW version 0.1.0"

// Button settings
#define BUTTON_DEBOUNCE_MS 500
#define BUTTON_REPEAT_MS 100
#define BUTTON_UP_MASK (1<<PB0)
#define BUTTON_DOWN_MASK (1<<PB1)
#define BUTTON_SEL_MASK (1<<PB2)

// ADC settings
#define FULL_RANGE_ADC_DEFAULT 802 // 802 calculated, 782 after cal/correction - built prototype
#define FULL_RANGE_VOLTAGE 5000 // using 120meg and 47k resistors, 2.5V reference
#define MAX_VOLTAGE 5000
#define ADJUST_STEP_DEFAULT 50
#define ADPS_2 0b001
#define ADPS_4 0b010
#define ADPS_8 0b011
#define ADPS_16 0b100
#define ADPS_32 0b101
#define ADPS_64 0b110
#define ADPS_128 0b111

// not writing REFSx bits as ext. ref is used
#define CS_MUX 0b0000
#define VS_MUX 0b0100
#define AD_OFFSET 0
#define AD_DISPLAY_AVG_PASSES 128
#define MIN_CURRENT_ADC_VAL_DISPLAY 5
#define OCP_RESET_I false
#define RSW_DELAY 1 // delay (averaging cycles) between next measurements and switching current shunt resistor

/*
 * DEFAULT PID SETTINGS
 * Warning, integral and derivative terms are scaled somewhat arbitrarily, depending on prescaler settings, and also referenced to the Kx values
 * Please check the algorithm/ISR itself, all values are rescaled to avoid using floating point maths
 * 
 * first increase KP with KI, KD=0 until unstable
 * then increase D until no oscillation
 * then increase I to solve steady state offset
 * then adjust all to get stable result
*/

#define KP_DEFAULT 5700U
#define KI_DEFAULT 13U
#define KD_DEFAULT 4500U

// Limits
#define LP_H_DEFAULT 2500U
#define LP_L_DEFAULT 2500U
#define LI_H_DEFAULT 55000U
#define LI_L_DEFAULT 45000U
#define LD_H_DEFAULT 8000U
#define LD_L_DEFAULT 8000U
// max contrib. to PWM: *Kx/4096

#define OVERSHOOT_PERCENT_PIDI_RESET 7 // percent of set value to be exceeded to reset I term
#define OVERSHOOT_RESERVE_PIDI_RESET 3 // + this reserve

#define PWM_LIMIT 220
#define OCP_RECOVERY_STEP 2
#define OCP_ATTACK_STEP 20

// LCD
#define RS 11 // PB3
#define E 12 // PB4
#define D4 4
#define D5 5
#define D6 6
#define D7 7

// Current sense inputs
#define SHUNT_2K2 A1 // PC1
#define SHUNT_22K A2 // PC2
#define SHUNT_220K A3 // PC3

// Button inputs
#define BTN1 8 // PB0
#define BTN2 9 // PB1
#define BTN3 10 // PB2

// General Purpose output
#define GP 13 // PB5

// Over-Current detection, BOD
#define PWRFAIL 0 // PD0
#define SEC_OC_DISABLE  1 // PD1
#define OC_INPUT 2 // PD2

// variables
uint16_t kp, ki, kd; // PID constants
uint16_t lpH, lpL, liH, liL, ldH, ldL; // PID limits
uint16_t fullRangeADC; // voltages, ranges
uint8_t adjustStep = 50; // voltage adjust step (TBD for user settings...)
volatile boolean pidChangeOrRead = false; // read AD value from PID (AD) interrupt or change set point
volatile boolean pwmEnable = false, pidEnable = true;
volatile uint8_t displayOCP = 0;
uint8_t loopRun = 0, pwmRead, pwm, maxPwm, prevPWM, currentReads = 0;
int32_t pidP = 0, pidI = 0, pidD = 0;
int16_t adcValue = 0, adcValueOld = 0, adcValueBuf = 0, pidSetPoint = 0, setPoint = 0, adcVoltRead, adcCurrRead, currentADC = 0;
uint16_t outputVoltage = 0, targetVoltage = 0, pidVoltage = 0;
uint32_t outputCurrent = 0; // *10nA
uint32_t adcVoltAveraging = 0;
uint32_t adcCurrAveraging = 0;
uint16_t adcAveragingCount = 0;
boolean waitForPid;
unsigned long buttonTimer = 0;
uint16_t PIResetOvershootValue = 0;
uint8_t currReady = 0; // current ready counter
uint8_t currOLUL = 0; // current over (1) or under limit (2)
uint8_t currSel = 0, prevCurrSel = 255; // current sense resistor choice (0: 220020, 1: 22020, 2: 2220 ohm)
uint8_t rswDelay = 0; // downcounting - delay counter after shunt resistor switch, see RSW_DELAY

// custom characters
uint8_t lcdSmallOmegaLetter[8] = {
  0,
  0,
  0b00001110,
  0b00010001,
  0b00010001,
  0b00001010,
  0b00011011,
  0
};

uint8_t lcdSmallMuLetter[8] = {
  0,
  0b00010001,
  0b00010001,
  0b00010001,
  0b00010011,
  0b00011101,
  0b00010000,
  0b00010000
};

LiquidCrystal lcd(RS, E, D4, D5, D6, D7);

uint8_t adcChannel;
boolean OCP = false; // over-current pulse

volatile boolean isrCurrentOL = false; // found current out of range in ISR? (for short overshoots)

// INT0 interrupt for OCP
ISR(INT0_vect) {
  OCP = true;
}

// synchronize ADC to PWM
volatile uint8_t evenCounter; // counter will trigger ADC on even Timer 2 ISR runs, therefore ~7.8 kHz PID speed
ISR(TIMER2_OVF_vect) {
  if (evenCounter == 1) {
    ADCSRA |= (1 << ADSC);
  }
  evenCounter++;
  evenCounter&=1;
}

// ADC interrupt.
// The PID loop is handled here
ISR(ADC_vect, ISR_NOBLOCK) {
  PORTB ^= (1<<PB5);
  ADMUX = 0b1111; // GND
  delayMicroseconds(1); // yes, interrupt in isr :( VERY dirty fix ~(,, ,,°> to prevent channels affecting each other

  if (~PIND & (1 << PD2)) OCP = true; // check if ~overcurrent input pulled down

  uint16_t ADread = ADC;

  if (adcChannel == 0) {
    ADMUX = CS_MUX;
    adcValueBuf = ADread;
    // save previous ADC value, read new one
    adcValueOld = adcValue;
    adcValue = adcValueBuf;

    // change/read code
    if (pidChangeOrRead) {
      pidSetPoint = setPoint;
      adcVoltRead = adcValue;
      adcCurrRead = currentADC;
      pwmRead = pwm;
      pidChangeOrRead = false;
    }

    // proportional, integral, derivative values (-error)
    pidP = ((int16_t)pidSetPoint - adcValue);
    pidI += pidP;
    pidD = (int16_t)adcValueOld - adcValue;

    // pin 10 = PB2 = Overcurrent detection, active low
    if (OCP) { // OCP triggered
      if (maxPwm >= OCP_ATTACK_STEP) maxPwm -= OCP_ATTACK_STEP;
      else maxPwm = 0;
      if (OCP_RESET_I && pidI > 0) pidI = 0;
      displayOCP = 2;
    } else { // OCP not triggered
      if (maxPwm <= (PWM_LIMIT - OCP_RECOVERY_STEP)) {
        maxPwm += OCP_RECOVERY_STEP;
      }
      else maxPwm = PWM_LIMIT;
    }

    if (pwmEnable == false) {
      pwm = 0;
      pidI = 0;
    }

    OCP = false; // OCP signal reset

    // bounds check
    if (pidP > (int32_t)lpH) pidP = (int32_t)lpH;
    if (pidP < -((int32_t)lpL)) pidP = -(int32_t)lpL;
    if (pidI > (int32_t)liH) pidI = (int32_t)liH;
    if (pidI < -((int32_t)liL)) pidI = -(int32_t)liL;
    if (pidD > (int32_t)ldH) pidD = (int32_t)ldH;
    if (pidD < -((int32_t)ldL)) pidD = -(int32_t)ldL;

    if ((int16_t)adcValue > pidSetPoint && ((adcValue - pidSetPoint) > (int16_t)PIResetOvershootValue)) pidI = 0;

    // PWM adjustment
    int32_t out = ((int32_t)pidP * (int32_t)kp) + ((int32_t)pidI * (int32_t)ki) + ((int32_t)pidD * (int32_t)kd);
    uint32_t outU = out > 0 ? ((uint32_t)out >> 12) : 0; // compare, convert to unsigned (dividing signed ints is slow) and divide by 4096

    if (outU < maxPwm) pwm = (uint8_t)outU;
    else pwm = maxPwm;
    OCR2B = pwm;
    loopRun++;
    adcValueBuf = 0;
  }

  else {
    ADMUX = VS_MUX;
    currentADC = ADread;
    if (currentADC > 980) isrCurrentOL = true;
    currentReads = (currentReads < 255 ? currentReads + 1 : currentReads);
  }

  adcChannel++; // selection of V/I Measurement
  adcChannel &= 1;

}

void setup() {
  pinMode(SHUNT_2K2, OUTPUT);
  pinMode(SHUNT_22K, OUTPUT);
  pinMode(SHUNT_220K, OUTPUT);
  pinMode(BTN1, INPUT_PULLUP);
  pinMode(BTN2, INPUT_PULLUP);
  pinMode(BTN3, INPUT_PULLUP);
  pinMode(GP, OUTPUT);
  pinMode(SEC_OC_DISABLE, OUTPUT);
  digitalWrite(SEC_OC_DISABLE, LOW);
  pinMode(OC_INPUT, INPUT_PULLUP);
  pinMode(PWRFAIL, INPUT_PULLUP);

  //setup LCD
  lcd.begin(16, 2);
  lcd.createChar(0, lcdSmallOmegaLetter);
  lcd.createChar(1, lcdSmallMuLetter);
  lcd.setCursor(0, 0);
  lcd.print(F("HV power supply"));
  lcd.setCursor(0, 1);
  lcd.print(F(FW_VER_STRING));

  pinMode(A5, OUTPUT);
  pinMode(10, INPUT_PULLUP);

  delay(20);

  if (readCalibrationEE() == false) {
    loadDefaults();
    lcd.setCursor(0, 0);
    lcd.print(F("EEPROM corrupted"));
    lcd.setCursor(0, 1);
    lcd.print(F("Using basic cfg."));
    delay(2000);
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.println(F("  Hold SEL for  "));
    lcd.setCursor(0, 1);
    lcd.println(F("  calibration.  "));
    delay(2000);
  }

  boolean enterCal = false;
  if ((~PINB) & BUTTON_SEL_MASK) enterCal = true;

  delay(500);
  lcd.clear();

  if (enterCal) {
    lcd.setCursor(0, 0);
    lcd.print(F("Hit SEL quickly"));
    lcd.setCursor(0, 1);
    lcd.print(F("for calibration"));
    boolean selBtn = false, prevSelBtn = false;
    uint8_t calEnterTimer = 240, pressCount = 0;
    while (calEnterTimer--) {
      selBtn = ((~PINB) & BUTTON_SEL_MASK);
      if (selBtn != prevSelBtn) pressCount++;
      prevSelBtn = selBtn;
      if (pressCount < 10) delay(25); // 6 sec
    }
    if (pressCount < 10) enterCal = false;
  }

  if (enterCal) {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print(F("Calibration mode"));
    lcd.setCursor(0, 1);
    lcd.print(F("Be careful!"));
    delay(1500);
    doCalibration();
  }

  lcd.clear();

  // configure peripherals
  configPeriph();

  // start generator
  pwmEnable = true;

  pidChangeOrRead = true;
}

uint8_t buttons, buttonPulse;
boolean buttonBlock = false, buttonRepeat = false;


#define DISPLAY_PREDIV AD_DISPLAY_AVG_PASSES

boolean modeVoltage = true;
boolean modeIsol = false;
boolean rampUpFinished = false;
char buf[20]; //for display

void loop() {
  
  if (loopRun % 8 == 0) {
    waitForPid = true;  // 31 kHz -> triggered at ~4.8 kHz rate
    pidChangeOrRead = true;
    
  }

  if (waitForPid == true && pidChangeOrRead == false) {
    adcVoltAveraging += adcVoltRead;
    adcCurrAveraging += adcCurrRead;
    adcAveragingCount++;
    if (adcAveragingCount == AD_DISPLAY_AVG_PASSES) {
      outputVoltage = ADtoVoltage(adcVoltAveraging, AD_DISPLAY_AVG_PASSES);
      outputCurrent = ADtoCurrent(adcCurrAveraging, AD_DISPLAY_AVG_PASSES, &currSel, &currReady, &rswDelay, &currOLUL, (boolean*)(&isrCurrentOL));
      adcAveragingCount = 0;
      adcVoltAveraging = 0;
      adcCurrAveraging = 0;
    }

    if (currSel != prevCurrSel) {
      DDRC &= ~((1 << PC3) | (1 << PC2) | (1 << PC1));
      switch (currSel) {
        case 0:
          DDRC |= (1 << PC3);
          break;
        case 1:
          DDRC |= (1 << PC2);
          break;
        case 2:
        default:
          DDRC |= (1 << PC1);
          break;
      }
    }
    prevCurrSel = currSel;

    if (rswDelay == 0 && currReady < 255) currReady++;

    // Display measured values after all averaging etc. is done
    if (adcAveragingCount % DISPLAY_PREDIV == 0) {
      if (modeVoltage) {
        lcd.setCursor(0, 0);
        lcd.print("Set: ");
        lcd.print(targetVoltage);
        lcd.print("V   ");
        lcd.setCursor(0, 1);
        lcd.print("ADC: ");
        lcd.print(outputVoltage);
        lcd.print("V   ");
        lcd.setCursor(13, 1);
        if (displayOCP > 0) {
          lcd.print("OCP");
          displayOCP--;
        }
        else lcd.print(" CV");
      } else if (modeIsol) { // Isolation/current/resistance mode
        lcd.setCursor(0, 0);
        lcd.print("Cur: ");
        if (currReady < 5) lcd.print (F("----          "));
        else {
          if (currOLUL == 0) {
            if (outputCurrent <= 2000ul) { // <= 2uA, res 1nA
              lcd.print((uint16_t) outputCurrent);
              lcd.print("nA      ");
            } else { // x uA, res 10/100nA/1uA
              if (outputCurrent <= 20000ul) {
                sprintf(buf, "%02u.%02u%cA     ", (uint16_t)(outputCurrent / 1000ul), (uint16_t)((outputCurrent % 1000ul) / 10u), 1); // 10nA res, <=20uA
                lcd.print(buf);
              } else if (outputCurrent <= 200000ul) {
                sprintf(buf, "%03u.%01u%cA     ", (uint16_t)(outputCurrent / 1000ul), (uint16_t)((outputCurrent % 1000ul) / 100u), 1); // 100nA res, <= 200uA
                lcd.print(buf);
              } else {
                lcd.print((uint16_t)(outputCurrent / 1000ul));
                lcd.print((char)(1));//mu letter
                lcd.print(F("A     "));
              }
            }
          } else if (currOLUL == 1) {
            lcd.print(F("Over limit   "));
          } else if (currOLUL == 2) {
            lcd.print(F("Under limit  "));
          }

        }
        lcd.setCursor(0, 1);
        lcd.print("Res: ");
        // Resistance calculation and display
        if (currReady < 5 || currOLUL > 0) lcd.print ("----        ");
        else if (outputCurrent < 50) lcd.print (F("Over limit  ")); // <50 nA would give nonsense result anyway
        else {
          uint32_t kohms = outputVoltage; // Ohm's law R=V/I
          if (outputVoltage > 2500) kohms *= 500000ul;
          else kohms *= 1000000ul;
          kohms /= (outputVoltage > 2500) ? outputCurrent / 2ul : outputCurrent;
          if (kohms < 500) lcd.print ("Under limit   "); // would give bad result
          else
          {
            kohms -= 10; // internal resistance cca 10kOhm
            if (kohms > 30000000ul) lcd.print (F("Over limit   "));
            else if (kohms > 1000000) {
              lcd.print(kohms / 1000000);
              lcd.print(".");
              sprintf(buf, "%03u", (uint16_t)((kohms % 1000000) / 1000));
              lcd.print(buf);
              lcd.print("G");
              lcd.print((char)0);//ohm
              lcd.print("   ");
            } else {
              lcd.print(kohms / 1000);
              lcd.print(".");
              sprintf(buf, "%03u", (uint16_t)(kohms % 1000));
              lcd.print(buf);
              lcd.print("M");
              lcd.print((char)0);//ohm
              lcd.print("   ");
            }
          }
        }


      }
    }

    waitForPid = false;

    // slowly ramping up set voltage
    if (targetVoltage < pidVoltage) pidVoltage = targetVoltage;
    else {
      if (targetVoltage < adjustStep) pidVoltage = targetVoltage;
      else if (pidVoltage < targetVoltage) {
        if (pidVoltage < targetVoltage - adjustStep) pidVoltage += adjustStep;
        else pidVoltage = targetVoltage;
      }
    }
    if (targetVoltage == pidVoltage) rampUpFinished = true; else rampUpFinished = false;

    setPoint = voltToADpoint(pidVoltage);
    PIResetOvershootValue = (1UL * setPoint * OVERSHOOT_PERCENT_PIDI_RESET / 100UL) + OVERSHOOT_RESERVE_PIDI_RESET;
  }

  // buttons are on PB0:2
  buttons = (~PINB) & 0b111;
  buttonPulse = 0;
  if (buttons != 0 && !buttonBlock) {
    buttonPulse = buttons;
    buttonBlock = true;
    buttonTimer = millis() + (buttonRepeat ? BUTTON_REPEAT_MS : BUTTON_DEBOUNCE_MS);
  } else if (buttonBlock) {
    if (millis() >= buttonTimer) buttonBlock = false;
    if (!(buttons & BUTTON_SEL_MASK)) buttonRepeat = true; // no fast repeat if SEL held
  } else if (buttons == 0) {
    buttonRepeat = false;
  }

  if (buttonPulse & BUTTON_SEL_MASK) {
    if (modeVoltage) {
      modeVoltage = false;
      modeIsol = true;
    } else {
      modeVoltage = true;
      modeIsol = false;
    }
    currReady = 0;
    displayOCP = 0;
    lcd.clear();
  }

  if (buttonPulse & BUTTON_UP_MASK) {
    if (modeVoltage && targetVoltage < MAX_VOLTAGE) targetVoltage += adjustStep;
  }

  if (buttonPulse & BUTTON_DOWN_MASK) {
    if (modeVoltage && targetVoltage >= adjustStep) targetVoltage -= adjustStep;
  }

  if (!digitalRead(PWRFAIL)) { // in case of power input issues, shut down output and 
    lcd.clear();
    lcd.print("Insufficient");
    lcd.setCursor(0,1);
    lcd.print("input voltage!");
    targetVoltage = 0;
    pidVoltage = 0;
    outputVoltage = 0;
    setPoint = 0;
    boolean pwmEnableOld = pwmEnable;
    pwmEnable = false;
    modeIsol = false;
    modeVoltage = true;
    pidChangeOrRead = true;
    while (!digitalRead(PWRFAIL));
    delay(1000);
    lcd.clear();
    pwmEnable = pwmEnableOld;
    pidChangeOrRead = true;
  } 
}
// end loop 


uint16_t voltToADpoint(uint16_t voltage) {
  uint32_t adValCalcTemp = voltage; // temporary
  adValCalcTemp *= (uint32_t)fullRangeADC;
  adValCalcTemp /= (uint32_t)FULL_RANGE_VOLTAGE;
  adValCalcTemp = (uint32_t)(AD_OFFSET + adValCalcTemp);
  if (voltage) return (uint16_t)adValCalcTemp;
  else return 0;
}

uint16_t ADtoVoltage(uint32_t inValue, uint16_t summedSampleCount)
{
  uint32_t voltCalcTemp = inValue; // temporary
  voltCalcTemp = (uint32_t)(voltCalcTemp > AD_OFFSET * summedSampleCount ? (voltCalcTemp - (AD_OFFSET * summedSampleCount)) : 0);
  voltCalcTemp *= FULL_RANGE_VOLTAGE;
  voltCalcTemp /= fullRangeADC;
  voltCalcTemp /= summedSampleCount;
  return (uint16_t)voltCalcTemp;
}


uint32_t ADtoCurrent(uint32_t inValue, uint16_t summedSampleCount, uint8_t* currSel, uint8_t* currReady, uint8_t* rswCounter, uint8_t* currOLUL, boolean* isrIOL) {
  *currOLUL = 0;
  uint16_t avgAdcValue = (uint16_t)(inValue / (uint32_t)summedSampleCount);

  if (*rswCounter == 0) { // after shunt R switched
    if (avgAdcValue > 900 || *isrIOL) {
      if (*currSel < 2) {
        (*currSel)++;
        *rswCounter = RSW_DELAY;
        *currReady = 0;
      } else {
        *currOLUL = 1; // Over Limit
      }
    } else if (avgAdcValue < 70 && *currSel > 0) {
      (*currSel)--;
      *currReady = 0;
      *rswCounter = RSW_DELAY;
    } else if (avgAdcValue < MIN_CURRENT_ADC_VAL_DISPLAY) {
      *currOLUL = 2; // Under Limit
    }
  } else (*rswCounter)--;
  *isrIOL = false;

  // Current is in nA
  uint32_t currCalcTemp = inValue; // temporary

  switch (*currSel) {
    case 0:// 220020 ohm
      currCalcTemp *= 1109ul; // 11.09 nA per LSB
      currCalcTemp /= 100ul;
      break;
    case 1: // 22020 ohm
      currCalcTemp *= 1109ul; // 110.9 nA per LSB
      currCalcTemp /= 10ul;
      break;
    default: // 2220 ohm
      currCalcTemp *= 1110ul; // 1110 nA per LSB
      break;
  }

  currCalcTemp /= (uint32_t)summedSampleCount;
  return currCalcTemp;
}

uint16_t chksum = 0, chksumEE = 0;

// HERE we're working with global variables inside functions for easy modability, yes dirty again
boolean readCalibrationEE() {
  EEPROM.get(0, fullRangeADC);
  EEPROM.get(2, kp);
  EEPROM.get(4, ki);
  EEPROM.get(6, kd);
  EEPROM.get(8, lpH);
  EEPROM.get(10, lpL);
  EEPROM.get(12, liH);
  EEPROM.get(14, liL);
  EEPROM.get(16, ldH);
  EEPROM.get(18, ldL);
  EEPROM.get(100, chksumEE);

  chksum = 0xaaaa ^ fullRangeADC ^ kp ^ ki ^ kd ^ lpH ^ lpL ^ liH ^ liL ^ ldH ^ ldL;
  if (chksum != chksumEE || kp == 65535u || ki == 65535u || kd == 65535u || kp == 0u || ki == 0u || kd == 0u) {
    return false; // if error
  }
  return true;
}

void loadDefaults() {
  fullRangeADC = FULL_RANGE_ADC_DEFAULT;
  adjustStep = ADJUST_STEP_DEFAULT;
  kp = KP_DEFAULT;
  ki = KI_DEFAULT;
  kd = KD_DEFAULT;
  lpH = LP_H_DEFAULT;
  lpL = LP_L_DEFAULT;
  liH = LI_H_DEFAULT;
  liL = LI_L_DEFAULT;
  ldH = LD_H_DEFAULT;
  ldL = LD_L_DEFAULT;
  chksum = 0xaaaa ^ fullRangeADC ^ kp ^ ki ^ kd ^ lpH ^ lpL ^ liH ^ liL ^ ldH ^ ldL;
}

// to adjust an uint16_t
uint8_t pressTime;

void adjustUint16(uint16_t* toAdjust, char* text, uint16_t lowLimit, uint16_t highLimit, uint16_t stepSize) {
  uint8_t adjRepTime = 6, pressTime = 0; // *50 ms before first repeat
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print(text);
  uint16_t previousToAdjust = *toAdjust + 1;
  uint8_t prevBtns = 0, btns = 0;
  boolean setting = true;
  while (setting) {
    if (*toAdjust != previousToAdjust) {
      lcd.setCursor(0, 1);
      lcd.print(*toAdjust);
      lcd.print(F("     "));
    }
    if (adjRepTime == pressTime) {
      prevBtns = 0; //simulate another press
      adjRepTime = 2; // *50ms before next repeat
      pressTime = 0;
    } else if (btns == 0) adjRepTime = 6; // back to normal time  before repeat
    btns = (~PINB) & (BUTTON_SEL_MASK | BUTTON_UP_MASK | BUTTON_DOWN_MASK);
    uint8_t buttonPress = btns & (btns ^ prevBtns);
    prevBtns = btns;

    if ((buttonPress & BUTTON_UP_MASK) && *toAdjust <= highLimit - stepSize) {
      (*toAdjust) += stepSize;
      while (((~PINB) & BUTTON_UP_MASK) && pressTime < adjRepTime) {
        delay(50);
        pressTime++;
      }
    }
    else if ((buttonPress & BUTTON_DOWN_MASK) && *toAdjust >= lowLimit + stepSize) {
      (*toAdjust) -= stepSize;
      while (((~PINB) & BUTTON_DOWN_MASK) && pressTime < adjRepTime) {
        delay(50); // dirty debounce
        pressTime++;
      }
    }
    else if (buttonPress & BUTTON_SEL_MASK) {
      lcd.clear();
      while ((~PINB) & BUTTON_SEL_MASK) delay(50);
      delay(150);
      setting = false;
    }

    delay(20);
  }
}

// needs to be done better, use flash for text and put limits in defines
void doCalibration() {
  lcd.clear();
  adjustUint16(&fullRangeADC, "Full Range ADC", 700, 900, 1);
  adjustUint16(&kp, "PID Kp", 10, 12000, 10);
  adjustUint16(&ki, "PID Ki", 1, 50, 1);
  adjustUint16(&kd, "PID Kd", 10, 12000, 10);
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print("Limits: *Kx/4096");
  delay(1500);
  lcd.clear();
  adjustUint16(&lpH, "P limit high", 0, 15000, 10);
  adjustUint16(&lpL, "P limit low (-)", 0, 15000, 10);
  adjustUint16(&liH, "I limit high", 0, 60000, 20);
  adjustUint16(&liL, "I limit low (-)", 0, 60000, 20);
  adjustUint16(&ldH, "D limit high", 0, 15000, 10);
  adjustUint16(&ldL, "D limit low (-)", 0, 15000, 10);
  saveEE();
}

// without delays, it freezes
void saveEE() {
  chksum = 0xaaaa ^ fullRangeADC ^ kp ^ ki ^ kd ^ lpH ^ lpL ^ liH ^ liL ^ ldH ^ ldL;
  EEPROM.put(0, fullRangeADC);
  delay(10);
  EEPROM.put(2, kp);
  delay(10);
  EEPROM.put(4, ki);
  delay(10);
  EEPROM.put(6, kd);
  delay(10);
  EEPROM.put(8, lpH);
  delay(10);
  EEPROM.put(10, lpL);
  delay(10);
  EEPROM.put(12, liH);
  delay(10);
  EEPROM.put(14, liL);
  delay(10);
  EEPROM.put(16, ldH);
  delay(10);
  EEPROM.put(18, ldL);
  delay(10);
  EEPROM.put(100, chksum);
  delay(10);
  lcd.print(F("Checksum: "));
  lcd.print(chksum);
  delay(2000);
}

void configPeriph() {
  // start ADC
  DDRC &= ~((1 << PC0) | (1 << PC4));
  PORTC &= ~0b00011111; // ADC shunts etc
  ADCSRA = ADPS_32 | (1 << ADEN) | (1 << ADIE); // set prescaler to :32 (38 kHz at 13 cycles), enable AD, enable interrupt
  ADMUX = VS_MUX; // set reference and input

  // enable INT0 falling edge interrupt
  EICRA = (1 << ISC01);
  EIMSK = (1 << INT0);

  // ~31.2 kHz pwm, non-inverting mode and overflow interrupt
  DDRD |= (1 << PD3);
  OCR2B = 0;
  TCCR2A = (1 << COM2B1) | (1 << WGM20);
  TCCR2B = (1 << CS20);
  TIMSK2 = (1 << TOIE2);

  delay(50);  
}


